"""
Build Cloud Formation stacks

"""
import logging

from formulary.builders import base
from formulary.resources import cloudformation
from formulary import config
from formulary.builders import rds
from formulary.builders import service
from formulary import utils

LOGGER = logging.getLogger(__name__)


class Stack(base.Builder):

    def __init__(self, config, name, base_path, amis, environment_stack):
        super(Stack, self).__init__(config, name)
        self._amis = amis
        self._base_path = base_path
        self._config = config
        self._mappings = config.mappings
        self._mappings.update(environment_stack.mappings)
        self._environment_stack = environment_stack
        self._process_resources()

    def _get_builder_config(self, resource, config_values):
        return config.BuilderConfig(config_values,
                                    self._mappings,
                                    self._config.region,
                                    self._config.s3_bucket,
                                    self._config.s3_prefix,
                                    self._config.profile,
                                    self._config.environment,
                                    resource)

    def _process_resources(self):
        for resource in self._config.settings.get('resources'):
            if resource['type'] == 'service':
                LOGGER.debug('Add service: %r', resource['name'])
                cfg_obj = config.ResourceConfig(self._base_path,
                                                resource['type'],
                                                resource['name'],
                                                self._config.environment)
                builder_cfg = self._get_builder_config(resource['name'],
                                                       cfg_obj.load())

                dependency = self._maybe_camel_case(resource.get('dependency'))
                handle = self._maybe_camel_case(resource.get('wait'))
                builder = service.Service(builder_cfg,
                                          resource['name'],
                                          self._amis,
                                          cfg_obj.resource_folder,
                                          self._environment_stack,
                                          dependency,
                                          handle,
                                          resource['name'])

                self._outputs += builder.outputs
                self._resources += builder.resources

            elif resource['type'] == 'rds':
                LOGGER.debug('Add RDS: %r', resource['name'])
                cfg_obj = config.ResourceConfig(self._base_path,
                                                resource['type'],
                                                resource['name'],
                                                self._config.environment)
                builder_cfg = self._get_builder_config(resource['name'],
                                                       cfg_obj.load())

                builder = rds.RDS(builder_cfg,
                                  resource['name'],
                                  self._environment_stack)

                self._outputs += builder.outputs
                self._resources += builder.resources

            elif resource['type'] == 'wait':
                LOGGER.debug('Add wait-condition: %r', resource['name'])
                if resource.get('handle'):
                    handle = {'Ref': utils.camel_case(resource['handle'])}
                    wait = cloudformation.WaitCondition(resource.get('count',1),
                                                        handle,
                                                        resource.get('timeout',
                                                                     3600))
                else:
                    wait = cloudformation.WaitCondition()
                    wait.set_creation_policy(resource['count'],
                                             resource['timeout'])
                if resource.get('dependency'):
                    wait.set_dependency(resource['dependency'])
                self._add_resource(resource['name'], wait)

                cc_name = utils.camel_case(resource['name'])
                self._add_output(cc_name + 'Data',
                                 'WaitCondition return data',
                                  {'Fn::GetAtt': [cc_name, 'Data']})

            elif resource['type'] == 'wait-handle':
                LOGGER.debug('Add wait handle: %s', resource['name'])
                self._add_resource(resource['name'],
                                   cloudformation.WaitConditionHandle())

            else:
                ValueError('Unsupported resource type: %s', resource['type'])

    @staticmethod
    def _maybe_camel_case(value):
        return utils.camel_case(value) if value else None



"""
Cloud Formation Template

"""
import collections
import json


class Template(object):
    """Used to create a Cloud Formation configuration template"""
    PARENT_CONFIG_PREFIX = ''
    STACK_TYPE = ''

    def __init__(self, name):
        """Create a new Cloud Formation configuration template instance"""
        self._description = 'Formulary created Cloud Formation stack'
        self._name = name

        self._conditions = {}
        self._mappings = {}
        self._metadata = {}
        self._outputs = []
        self._parameters = []
        self._resources = []

    def add_output(self, name, description, reference):
        """Add an output to the template

        :param str name: The name of the output
        :param str description: The description of the output
        :param str reference: The object to reference

        """
        self._outputs.append((name, {'Description': description,
                                     'Value': {'Ref': reference}}))

    def add_resource(self, resource_id, resource):
        """Add a resource to the template

        :param str resource_id: The camelCase resource_id
        :param Resource resource: The resource to add

        """
        self._resources.append((resource_id, resource))

    def as_json(self, indent=2):
        """Return the cloud-formation template as JSON

        :param int indent: spaces to indent the JSON by

        """
        value = collections.OrderedDict([
            ('AWSTemplateFormatVersion', '2010-09-09'),
            ('Conditions', self._conditions),
            ('Description', self._description),
            ('Mappings', self._mappings),
            ('Metadata', self._metadata),
            ('Outputs', collections.OrderedDict(self._outputs)),
            ('Parameters', collections.OrderedDict(self._parameters)),
            ('Resources', collections.OrderedDict(self._resources))])

        return json.dumps(value, indent=indent, sort_keys=False)

    @property
    def name(self):
        """Return the template name

        :rtype: str

        """
        return self._name

    def set_description(self, description):
        """Set the template description

        :param str description: The description value
        """
        self._description = description

    def update_mappings(self, mappings):
        """Update the mappings with values from another dict.

        :param dict mappings: Additional mappings to add

        """
        self._mappings.update(mappings)

    def update_outputs(self, outputs):
        """Update the outputs with values from another dict.

        :param list outputs: Additional outputs to add

        """
        self._outputs += outputs

    def update_parameters(self, parameters):
        """Update the parameters with values from another dict.

        :param list parameters: Dict of parameters to merge in

        """
        self._parameters += parameters

    def update_resources(self, resources):
        """Update the resources with the resources from another dict.

        :param list resources: Additional resources to add

        """
        self._resources += resources



"""
Common Utility Methods

"""
import re

DEFAULT_PROTOCOL = 'tcp'


def camel_case(value):
    """Convert a underscore or dash delimited value to camelCase

    :param str value: The value to convert
    :rtype: str

    """
    return ''.join(x.capitalize() for x in value.replace('-', '_').split('_'))


def dash_delimited(value):
    """Convert a camelCase value to a dash-delimited value

    :param str value: The value to convert
    :rtype: str

    """
    return re.sub('([a-z0-9])([A-Z])', r'\1-\2',
                  re.sub('(.)([A-Z][a-z]+)', r'\1-\2', value)).lower()


def find_in_map(value):
    """If the value is a ^map macro, replace it with a Cloud Formation
    Formation ``Fn::FindInMap`` function, splitting up the period delimited
    reference. For example if a value is ``^map Foo.Bar.Baz`` it will be
    returned as ``dict({'Fn::FindInMap': ['Foo', 'Bar', Baz'})``.

    :param str value: The value to check for the map macro
    :rtype: str|dict
    :raises: ValueError

    """
    if value.startswith('^map '):
        ref = value[5:].split('.')
        if len(ref) != 3:
            raise ValueError('Invalid map reference: {}'.format(value[5:]))
        return {'Fn::FindInMap': ref}
    return value


def parse_port_value(value, default_protocol=None):
    """Parse a string containing port information and return a normalized
    tuple of ``protocol``, ``from_port``, and ``to_port``.

    :param str|int value: The value to parse
    :return: str, int, int

    """
    if isinstance(value, int):
        return default_protocol or DEFAULT_PROTOCOL, value, value

    protocol = default_protocol or DEFAULT_PROTOCOL
    from_port, to_port = value, value

    if '-' in value:
        from_port, to_port = value.split('-')
        from_protocol, to_protocol = None, None
        if '/' in from_port:
            from_port, from_protocol = from_port.split('/')
        if '/' in to_port:
            to_port, to_protocol = to_port.split('/')
        protocol = (from_protocol or to_protocol or
                    default_protocol or DEFAULT_PROTOCOL)

    elif '/' in value:
        port, protocol = value.split('/')
        from_port, to_port = port, port

    if isinstance(protocol, str):
        protocol = protocol.lower()

    return protocol, int(from_port), int(to_port)



"""
Cloud Formation Port 53 Resources

"""
from formulary.resources import base


class Route53AliasTarget(base.Property):
    def __init__(self, dns_name, hosted_zone_id, eval_target_health=False):
        super(Route53AliasTarget, self).__init__()
        self._values = {'DNSName': dns_name,
                        'HostedZoneId': hosted_zone_id,
                        'EvaluateTargetHealth': eval_target_health}


class Route53HostedZone(base.Resource):

    tags = False

    def __init__(self, config, tags, name, vpcs):
        super(Route53HostedZone, self).__init__('AWS::Route53::HostedZone')
        self._properties['HostedZoneConfig'] = config
        self._properties['HostedZoneTags'] = tags
        self._properties['Name'] = name
        self._properties['VPCs'] = vpcs


class Route53HostedZoneConfig(base.Property):
    def __init__(self, comment):
        super(Route53HostedZoneConfig, self).__init__()
        self._values = {'Comment': comment}


class Route53HostedZoneTags(base.Property):
    def __init__(self, tags):
        super(Route53HostedZoneTags, self).__init__()
        self._values = [{'Key': k, 'Value': v} for k,v in tags]


class Route53HostedZoneVPCs(base.Property):
    def __init__(self, vpcs):
        super(Route53HostedZoneVPCs, self).__init__()
        self._values = [{'VPCId': k, 'VPCRegion': v} for k,v in vpcs]


class Route53RecordSet(base.Resource):

    tags = False

    def __init__(self, domain_name, hostname, resources=None,
                 alias_target=None, record_type='A', ttl=300, comment=None):
        super(Route53RecordSet, self).__init__('AWS::Route53::RecordSet')
        if alias_target and resources:
            raise ValueError('Can not have both resources and an alias target')
        if alias_target and ttl:
            ttl = None
        if not alias_target and not resources:
            raise ValueError('Must have either an alias target or resources')
        domain_name = domain_name.rstrip('.') + '.'
        self._properties['Comment'] = comment
        self._properties['AliasTarget'] = alias_target
        self._properties['Failover'] = None
        self._properties['GeoLocation'] = None
        self._properties['HealthCheckId'] = None
        self._properties['HostedZoneName'] = domain_name
        self._properties['Name'] = '{0}.{1}'.format(hostname, domain_name)
        self._properties['Region'] = None
        self._properties['ResourceRecords'] = resources
        self._properties['SetIdentifier'] = None
        self._properties['TTL'] = str(ttl) if ttl else None
        self._properties['Type'] = record_type
        self._properties['Weight'] = None


class Route53RecordSetGroup(base.Resource):

    tags = False

    def __init__(self, domain_name, record_sets, comment=None):
        super(Route53RecordSetGroup,
              self).__init__('AWS::Route53::RecordSetGroup')
        self._properties['Comment'] = comment
        self._properties['HostedZoneName'] = domain_name
        self._properties['RecordSets'] = record_sets



"""
Upload to and remove stacks from S3

"""
import datetime
import logging

from boto3 import session
from botocore import exceptions

LOGGER = logging.getLogger(__name__)
TTL = 3600


class S3(object):
    """The S3 class is used to upload and delete stack files from AWS S3"""

    def __init__(self, bucket_name, root_path, profile):
        """Create a new instance of the S3 class passing in the bucket name
        and the root path for all files uploaded.

        :param str bucket_name: The name of the bucket for the files
        :param str root_path: The root path for all of the files

        """
        self._bucket_name = bucket_name
        self._root_path = root_path
        self._session = session.Session(profile_name=profile)
        self._client = self._session.client('s3')
        self._resource = self._session.resource('s3')

    def delete(self, file_name):
        """Delete a file from S3 by name

        :param str file_name: The file to delete

        """
        obj = self._resource.Object(self._bucket_name, self._key(file_name))
        obj.delete()
        LOGGER.debug('Deleted %s', file_name)

    def fetch(self, file_name):
        """Retrieve a file from S3 by name

        :param str file_name: The file to fetch
        :return: bytes

        """
        obj = self._resource.Object(self._bucket_name, self._key(file_name))
        response = obj.get()
        return response['Body']

    def upload(self, file_name, content):
        """Upload a file to S3, returning a pre-signed URL for accessing
        the file.

        :param str file_name: The name of the file
        :param bytes content: The file content
        :rtype: str

        """
        key = self._key(file_name)
        expiration = (datetime.datetime.utcnow() +
                      datetime.timedelta(seconds=TTL)).isoformat()
        try:
            self._client.put_object(ACL='authenticated-read',
                                    Body=content,
                                    Bucket=self._bucket_name,
                                    Expires=expiration,
                                    Key=key)
        except exceptions.ClientError as error:
            LOGGER.warning('Error putting value in bucket: %s', error)
            if error.response['Error']['Code'] == 'NoSuchBucket':
                self._create_bucket()
                return self.upload(file_name, content)
            raise error
        LOGGER.debug('Uploaded %s (%i bytes)', file_name, len(content))
        return self._client.generate_presigned_url(
            'get_object', Params={'Bucket': self._bucket_name, 'Key': key})

    def _create_bucket(self):
        self._client.create_bucket(Bucket=self._bucket_name)
        LOGGER.debug('Created bucket: %s', self._bucket_name)

    def _key(self, file_name):
        """Return the key to a file, prepending the root path to the key.

        :param str file_name: The name of the file
        :rtype: str

        """
        return '{0}/{1}'.format(self._root_path, file_name)



"""
Main Formulary Controller

"""
import logging
import sys
import uuid

from os import path

from formulary.builders import environment
from formulary.builders import rds
from formulary.builders import service
from formulary.builders import stack
from formulary import cloudformation
from formulary import config
from formulary import template

LOGGER = logging.getLogger(__name__)

ACTIONS = {'create', 'update'}
RESOURCE_TYPES = {'environment', 'service', 'elasticache', 'rds',  'stack'}


class Controller(object):
    """The controller implements the top-level application behavior"""

    def __init__(self, config_path, action, env, resource_type, resource,
                 verbose, dry_run, profile):
        self._config_obj = config.ResourceConfig(config_path,
                                                 resource_type,
                                                 resource,
                                                 env)

        self._validate_arguments(config_path, action, env,
                                 resource_type, resource)

        self._action = action
        self._s3_prefix = str(uuid.uuid4())
        LOGGER.debug('S3 prefix: %s', self._s3_prefix)
        self._config_path = config_path
        self._environment = env
        self._profile = profile
        self._resource = resource
        self._resource_type = resource_type
        self._verbose = verbose
        self._dry_run = dry_run

        self._amis = self._config_obj.load_file('.', 'amis')
        self._instances = self._config_obj.load_file('.', 'instances')
        self._config = self._config_obj.load()
        self._environment_config = self._config_obj.environment_config()
        self._s3_bucket = self._environment_config.get('s3bucket')
        if not self._s3_bucket:
            self._s3_bucket = self._config.get('s3bucket')

        LOGGER.debug('S3bucket: %s', self._s3_bucket)

        if not self._s3_bucket:
            raise ValueError('s3bucket not configured in environment')

        self._mappings = self._config_obj.mappings()
        self._template = template.Template(self._template_name)
        self._environment_stack = self._get_environment_stack()
        self._cloud_formation = cloudformation.CloudFormation(self._profile,
                                                              self._region,
                                                              self._s3_bucket,
                                                              self._s3_prefix)

    def execute(self):
        """Create or update a Cloud Formation stack"""
        self._build_template_resources()

        if self._config.get('description'):
            self._template.set_description(self._config['description'])

        template_value = self._template.as_json()

        if self._dry_run:
            print(template_value)
            return

        env = self._resource if not self._environment else self._environment
        service_name = self._resource if self._environment else None

        if self._action == 'create':
            try:
                stack_id = self._cloud_formation.create_stack(self._template,
                                                              env,
                                                              service_name)
                print('Stack {0} created'.format(stack_id))
            except cloudformation.RequestException as error:
                self._error(str(error))

        elif self._action == 'update':
            try:
                self._cloud_formation.update_stack(self._template)
                print('Stack updated')
            except cloudformation.RequestException as error:
                self._error(str(error))

    def _build_environment_resources(self):
        builder_config = config.BuilderConfig(self._config, self._mappings,
                                              self._region, self._s3_bucket,
                                              self._s3_prefix, self._profile)
        builder = environment.Environment(builder_config, self._resource)
        self._template.update_outputs(builder.outputs)
        self._template.update_parameters(builder.parameters)
        self._template.update_resources(builder.resources)

    def _build_rds_resources(self):
        builder_config = self._get_builder_config()

        if builder_config.settings.get('stack-only'):
            self._error("The specified service is designated as stack-only.")

        builder = rds.RDS(builder_config, self._resource,
                          self._environment_stack)
        self._template.update_outputs(builder.outputs)
        self._template.update_parameters(builder.parameters)
        self._template.update_resources(builder.resources)

    def _build_service_resources(self):
        service_path = path.join(self._config_path,
                                 self._config_obj.resource_folder)
        builder_config = self._get_builder_config()

        if builder_config.settings.get('stack-only'):
            self._error("The specified service is designated as stack-only.")

        builder = service.Service(builder_config,
                                  self._resource, self._amis,
                                  service_path, self._environment_stack)
        self._template.update_outputs(builder.outputs)
        self._template.update_parameters(builder.parameters)
        self._template.update_resources(builder.resources)
        return builder

    def _build_stack_resources(self):
        builder = stack.Stack(self._get_builder_config(),
                              self._resource, self._config_obj.base_path,
                              self._amis, self._environment_stack)
        self._template.update_outputs(builder.outputs)
        self._template.update_parameters(builder.parameters)
        self._template.update_resources(builder.resources)
        return builder

    def _build_template_resources(self):
        self._template.update_mappings(self._mappings)
        if self._resource_type == 'environment':
            return self._build_environment_resources()

        self._template.update_mappings(self._environment_stack.mappings)

        if self._resource_type == 'rds':
            return self._build_rds_resources()
        elif self._resource_type == 'service':
            return self._build_service_resources()
        elif self._resource_type == 'stack':
            return self._build_stack_resources()

    def _get_builder_config(self):
        return config.BuilderConfig(self._config, self._mappings,
                                    self._region, self._s3_bucket,
                                    self._s3_prefix, self._profile,
                                    self._environment, self._resource)

    @staticmethod
    def _error(message):
        """Write out an error message and exit.

        :param str message: The error message

        """
        sys.stderr.write('ERROR: {0}\n'.format(message))
        sys.exit(1)

    def _get_environment_stack(self):
        """Return a stack instance if the environment is not

        """
        if self._resource_type == 'environment':
            return
        return cloudformation.EnvironmentStack(self._environment,
                                               self._environment_config,
                                               None, self._profile)

    @property
    def _region(self):
        """Return the region value from the environment config,
        defaulting to us-east-1

        :rtype: str

        """
        return self._environment_config.get('region', 'us-east-1')

    @property
    def _template_name(self):
        """Return the template name based upon the

        :rtype: str

        """
        if self._environment:
            return '{0}-{1}-{2}'.format(self._environment, self._resource_type,
                                        self._resource)
        return self._resource

    def _validate_arguments(self, config_path, action, env,
                            resource_type, resource):
        """Validate the initialization arguments, raising ``ValueError`` if any
        do not validate.

        :raises: ValueError

        """
        if not self._validate_action(action):
            raise ValueError('Invalid action: {0}'.format(action))

        if not self._validate_resource_type(resource_type):
            raise ValueError('Invalid resource type: {0}'.format(resource_type))

        if not self._config_obj.validate_config_path(config_path):
            raise ValueError('Invalid config path: {0}'.format(config_path))

        if resource_type != 'environment':
            if not env:
                raise ValueError('Environment not specified')
            elif not self._config_obj.validate_environment(config_path, env):
                raise ValueError('Invalid environment: {0}'.format(env))

        if not self._config_obj.validate_resource(config_path,
                                                  resource_type,
                                                  resource):
            raise ValueError('Invalid resource: {0}'.format(resource))

    @staticmethod
    def _validate_action(action):
        """Validate the action is a valid action

        :param str action: The action to validate
        :rtype: bool

        """
        return action in ACTIONS

    @staticmethod
    def _validate_resource_type(resource_type):
        """Validate the resource is valid

        :param str resource_type: The resource type specified
        :rtype: bool

        """
        return resource_type in RESOURCE_TYPES



"""
Cloud Formation Resources

AWS::ElastiCache::CacheCluster
AWS::ElastiCache::ParameterGroup
AWS::ElastiCache::ReplicationGroup
AWS::ElastiCache::SecurityGroup
AWS::ElastiCache::SecurityGroupIngress
AWS::ElastiCache::SubnetGroup

"""
from formulary.resources import base


class CacheCluster(base.Resource):
    """The AWS::ElastiCache::CacheCluster type creates an Amazon ElastiCache
    cache cluster.

    """
    def __init__(self):
        super(CacheCluster, self).__init__('AWS::ElastiCache::CacheCluster')
        self._properties = {'AutoMinorVersionUpgrade': None,
                            'AZMode': '',
                            'CacheNodeType': '',
                            'CacheParameterGroupName': '',
                            'CacheSecurityGroupNames': [],
                            'CacheSubnetGroupName': '',
                            'ClusterName': '',
                            'Engine': '',
                            'EngineVersion': '',
                            'NotificationTopicArn': '',
                            'NumCacheNodes': '',
                            'Port': 11211,
                            'PreferredAvailabilityZone': '',
                            'PreferredAvailabilityZones': [],
                            'PreferredMaintenanceWindow': '',
                            'SnapshotArns': [],
                            'SnapshotName': '',
                            'SnapshotRetentionLimit': 0,
                            'SnapshotWindow': '',
                            'VpcSecurityGroupIds': ['']}



"""
Cloud Formation ELB Resources

AWS::ElasticLoadBalancing::LoadBalancer

"""
from formulary.resources import base


class AccessLoggingPolicy(base.Property):
    """The AccessLoggingPolicy property describes where and how access logs are
    stored for the AWS::ElasticLoadBalancing::LoadBalancer resource.

    """
    def __init__(self, emit_interval, enabled, s3_bucket, s3_prefix):
        super(AccessLoggingPolicy, self).__init__()
        self._values = {'EmitInterval': emit_interval,
                        'Enabled': enabled,
                        'S3BucketName': s3_bucket,
                        'S3BucketPrefix': s3_prefix}


class AppCookieStickinessPolicy(base.Property):
    """The AppCookieStickinessPolicy type is an embedded property of the
    AWS::ElasticLoadBalancing::LoadBalancer type.

    """
    def __init__(self, cookie_name, policy_name):
        super(AppCookieStickinessPolicy, self).__init__()
        self._values = {'CookieName': cookie_name,
                        'PolicyName': policy_name}


class ConnectionDrainingPolicy(base.Property):
    """The ConnectionDrainingPolicy property describes how deregistered or
    unhealthy instances handle in-flight requests for the
    AWS::ElasticLoadBalancing::LoadBalancer resource. Connection draining
    ensures that the load balancer completes serving all in-flight requests
    made to a registered instance when the instance is deregistered or becomes
    unhealthy. Without connection draining, the load balancer closes
    connections to deregistered or unhealthy instances, and any in-flight
    requests are not completed.

    """
    def __init__(self, enabled, timeout):
        super(ConnectionDrainingPolicy, self).__init__()
        self._values = {'Enabled': enabled, 'Timeout': timeout}


class ConnectionSettings(base.Property):
    """ConnectionSettings is a property of the
    AWS::ElasticLoadBalancing::LoadBalancer resource that describes how long
    the front-end and back-end connections of your load balancer can remain
    idle.

    """
    def __init__(self, idle_timeout):
        super(ConnectionSettings, self).__init__()
        self._values = {'IdleTimeout': idle_timeout}


class HeathCheck(base.Property):
    """The ElasticLoadBalancing HealthCheck is an embedded property of the
    AWS::ElasticLoadBalancing::LoadBalancer type.

    """
    def __init__(self, interval=30, target='HTTP:80', timeout=5,
                 healthy=10, unhealthy=2):
        super(HeathCheck, self).__init__()
        self._values = {'HealthyThreshold': healthy,
                        'Interval': interval,
                        'Target': target,
                        'Timeout': timeout,
                        'UnhealthyThreshold': unhealthy}


class LBCookieStickinessPolicy(base.Property):
    """The LBCookieStickinessPolicy type is an embedded property of the
    AWS::ElasticLoadBalancing::LoadBalancer type.

    """
    def __init__(self, expiration_period, policy_name):
        super(LBCookieStickinessPolicy, self).__init__()
        self._values = {'CookieExpirationPeriod': expiration_period,
                        'PolicyName': policy_name}


class Listener(base.Property):
    """The Listener property is an embedded property of the
    AWS::ElasticLoadBalancing::LoadBalancer type.

    """
    def __init__(self, port, protocol, instance_port, instance_protocol,
                 policy_names=None, ssl_certificate_id=None):
        super(Listener, self).__init__()
        self._values = {'InstancePort': instance_port,
                        'InstanceProtocol': instance_protocol,
                        'LoadBalancerPort': port,
                        'PolicyNames':  policy_names or [],
                        'Protocol': protocol,
                        'SSLCertificateId': ssl_certificate_id}


class LoadBalancer(base.Resource):
    """The AWS::ElasticLoadBalancing::LoadBalancer type creates a LoadBalancer.

    """
    def __init__(self, name, instances, health_check, listeners,
                 security_groups, subnets,
                 availability_zone=None,
                 policies=None, cross_zone=False, internal=False):
        super(LoadBalancer,
              self).__init__('AWS::ElasticLoadBalancing::LoadBalancer')
        scheme = 'internal' if internal else 'internet-facing'
        self._properties = {'AvailabilityZone': availability_zone,
                            'CrossZone': cross_zone,
                            'HealthCheck': health_check.as_dict(),
                            'LoadBalancerName': name,
                            'Instances': instances,
                            'Listeners': [l.as_dict() for l in listeners],
                            'Policies': policies,
                            'Scheme': scheme,
                            'SecurityGroups': security_groups,
                            'Subnets': subnets}


class Policy(base.Property):
    """The ElasticLoadBalancing policy type is an embedded property of the
    AWS::ElasticLoadBalancing::LoadBalancer resource. You associate policies
    with a listener by referencing a policy's name in the listener's
    PolicyNames property.

    """
    def __init__(self, attributes, instance_ports, load_balancer_ports,
                 policy_name, policy_type):
        super(Policy, self).__init__()
        self._values = {'Attributes': attributes,
                        'InstancePorts': instance_ports,
                        'LoadBalancerPorts': load_balancer_ports,
                        'PolicyName': policy_name,
                        'PolicyType': policy_type}



"""
Base Builder Class

"""
import uuid

from formulary import s3
from formulary import template
from formulary import utils
from formulary.resources import cloudformation


class Builder(object):
    """Build CloudFormation stack templates"""
    def __init__(self, config, name):
        """Create a new instance of the stack builder

        :param fomulary.builders.Config config: Stack configuration
        :param str name: The name of the builder/stack resource

        """
        self._config = config
        self._environment = config.environment
        self._name = name
        self._outputs = []
        self._parameters = []
        self._resources = []
        self._templates = []

    def add_output(self, name, description, value):
        """Add an output that will be added to the template

        :param str name: The name of the output
        :param str|dict value: The value of the output

        """
        return self._add_output(name, description, value)

    def add_parameter(self, name, value):
        """Add a parameter that will be added to the template

        :param str name: The name of the parameter
        :param str|dict value: The value of the parameter

        """
        return self._add_parameter(name, value)

    def add_resource(self, name, resource):
        """Add a resource to the template, returning the cloud formation
        template reference name for the resource.

        :param str name: The underscore delimited resource name
        :param Resource resource: The resource to add
        :rtype: str

        """
        return self._add_resource(name, resource)

    @property
    def environment(self):
        """Return the environment value

        :rtype: str

        """
        return self._environment

    @property
    def full_name(self):
        """Return the builder's full name

        :rtype: str

        """
        if '-' in self.name:
            return self.name
        return '-'.join([self._config.service, self.name])

    @property
    def mappings(self):
        """Return the mappings dictionary for the builder

        :rtype: dict

        """
        return self._config.mappings

    @property
    def name(self):
        """Return the builder's resource name

        :rtype: str

        """
        return self._name

    @property
    def outputs(self):
        """Return the outputs dictionary for the builder

        :rtype: list

        """
        return self._outputs

    @property
    def parameters(self):
        """Return the parameters dictionary for the builder

        :rtype: list

        """
        return self._parameters

    @property
    def logical_id(self):
        return self.name

    @property
    def reference_id(self):
        """Return the camel case reference ID for the builder

        :rtype: str

        """
        return utils.camel_case(self.name)

    @property
    def resources(self):
        """Return the resource dictionary for the builder

        :rtype: list

        """
        return [(k, v.as_dict() if not isinstance(v, dict) else v) for k,v in self._resources]

    def upload(self, owner):
        """Upload a template for this builder to S3, returning the ID and
        the URL to the template

        :rtype: str, str

        """
        stack = template.Template(self._name)
        stack.set_description('Nested {0} stack owned by the '
                              '{1}-{2} stack'.format(self.__class__.__name__,
                                                     self.environment, owner))
        stack.update_mappings(self._config.mappings)
        stack.update_outputs(self.outputs)
        stack.update_parameters(self.parameters)
        stack.update_resources(self.resources)
        value = stack.as_json()
        template_id = str(uuid.uuid4())

        s3client = s3.S3(self._config.s3_bucket,
                         self._config.s3_prefix,
                         self._config.profile)
        return template_id, s3client.upload(template_id, value)

    def _add_output(self, name, description, value):
        """Add an output that will be added to the template

        :param str name: The name of the output
        :param str description: The description of the output
        :param str|dict value: The value of the output

        """
        output = {'Description': description, 'Value': value}
        self._outputs.append((name, output))

    def _add_parameter(self, name, value):
        """Add an parameter that will be added to the template

        :param str name: The name of the parameter
        :param str|dict value: The value of the parameter

        """
        self._parameters.append((name, value))

    def _add_resource(self, name, resource):
        """Add a resource to the template, returning the cloud formation
        template reference name for the resource.

        :param str name: The underscore delimited resource name
        :param Resource resource: The resource to add
        :rtype: str

        """
        resource_id = utils.camel_case(name)
        self._resources.append((resource_id, resource))
        return resource_id

    def _add_stack(self, name, template_url, parameters=None,
                   timeout=None, notifications=None, dependency=None):
        """Add a stack

        :param str template_url: URL to the stack to add
        :param dict parameters: Parameters to pass into the stack
        :param int timeout: Time out duration in minutes
        :param list notifications: A list of notification ARNs for the stack

        """
        stack = cloudformation.Stack(template_url, parameters,
                                     notifications, timeout)
        if dependency:
            stack.set_dependency(dependency)
        self._add_resource(name, stack)

    def _add_tag_to_resources(self, tag, value):
        for (k, resource) in self._resources:
            resource.add_tag(tag, value)

    def _maybe_replace_with_mapping(self, value):
        """If the value is a ^map macro, replace the with the value from the
        mappings dict. For example, if the value is ``^map Foo.Bar.Baz``
        return self._mappings['Foo']['Bar']['Baz'].

        :param str value: The value to check for the map macro
        :rtype: str|dict
        :raises: ValueError

        """
        if value.startswith('^map '):
            ref = value[5:].split('.')
            if len(ref) != 3:
                raise ValueError('Invalid map reference: {}'.format(value[5:]))
            return self._config.mappings[ref[0]][ref[1]][ref[2]]
        return value

    @property
    def _tags(self):
        return {'Environment': self._config.environment,
                'Service': self._config.service}



"""
Builder Configuration

"""
import logging
from os import path
import yaml

LOGGER = logging.getLogger(__name__)

CONFIG_FILES = {'amis.yaml', 'mapping.yaml', 'instances.yaml',
                'environments', 'rds', 'services'}

STACK_FOLDERS = {'elasticache': 'elasticache',
                 'environment': 'environments',
                 'rds': 'rds',
                 'service': 'services',
                 'stack': 'stacks'}


class BuilderConfig(object):

    """Configuration class for Builder objects"""
    def __init__(self, settings, mappings, region, s3_bucket, s3_prefix,
                 profile, environment=None, service=None):
        """Create a new instance of a Config class

        :param dict settings: Settings from configuration files
        :param dict mappings: Mapping configuration
        :param str region: AWS region name
        :param str s3_bucket: The formulary S3 work bucket
        :param str s3_prefix: s3 work prefix
        :param str profile: The AWS credentials profile to use
        :param str|None environment: The optional formulary environment name
        :param str|None service:  Formulary service name if set

        """
        self.settings = self._flatten_settings(settings, environment)
        self.environment = environment
        self.mappings = mappings
        self.region = region
        self.profile = profile
        self.s3_bucket = s3_bucket
        self.s3_prefix = s3_prefix
        self.service = service

    def _flatten_settings(self, settings, environment):
        out = {}
        for key, value in settings.items():
            if isinstance(value, dict):
                for value_key in settings[key].keys():
                    if value_key == environment:
                        out[key] = settings[key][value_key]
                        break
                    elif value_key == 'default':
                        out[key] = settings[key]['default']
                        break
                    else:
                        out[key] = self._flatten_settings(value, environment)
            else:
                out[key] = value
        return out


class ResourceConfig(object):

    def __init__(self, base_path, resource_type, resource, environment):
        self.base_path = self._normalize_path(base_path)
        self.environment = environment
        self._resource = resource
        self._resource_type = resource_type

    def _flatten_config(self, cfg):
        """Take a given config dictionary and if it contains environment
        specific values, map the environment values to the associated
        top level keys.

        :param dict cfg: The configuration to flatten
        :rtype: dict

        """
        output = {}
        for key, value in cfg.items():
            if isinstance(value, dict):
                if self.environment in value.keys():
                    output[key] = value[self.environment]
                else:
                    output[key] = self._flatten_config(value)
            elif isinstance(value, list):
                output[key] = []
                for list_value in value:
                    if isinstance(list_value, dict):
                        output[key].append(self._flatten_config(list_value))
                    else:
                        output[key].append(list_value)
            else:
                output[key] = value
        return output

    def load(self):
        """Return the config for the specified resource type

        :rtype: dict

        """
        if self._resource_type in ['environment', 'service']:
            settings = self.load_file(self.resource_folder, self._resource_type)
        else:
            settings = self.load_file(self.resource_folder, self._resource)
        return self._flatten_config(settings)

    def load_file(self, folder, file):
        """Return the contents of the specified configuration file in the
        specified configuration folder.


        :param str folder: The folder to load the configuration file from
        :param str file: The file to load
        :rtype: dict

        """
        file_path = path.join(self.base_path, folder, '{0}.yaml'.format(file))
        if path.exists(file_path):
            LOGGER.debug('Loading configuration from %s', file_path)
            with open(file_path, 'r') as handle:
                return yaml.load(handle)
        LOGGER.debug('Configuration file not found: %s', file_path)
        return {}

    def environment_config(self):
        """Return the environment configuration

        :rtype: dict

        """
        if not self.environment:
            return {}
        return self.load_file(path.join(STACK_FOLDERS['environment'],
                                        self.environment), 'environment')

    def environment_mappings(self):
        """Return the mappings from the environment folder

        :rtype: dict

        """
        if not self.environment:
            return {}
        return self.load_file(path.join(STACK_FOLDERS['environment'],
                                        self.environment), 'mappings')

    def mappings(self):
        """Return the mapping data from the various config dirs and return
        merged mapping values in order of precedence of global, environment,
        or service/entity.

        :rtype: dict

        """
        mappings = dict()
        mappings.update(self.load_file(self.base_path, 'mappings'))
        if not self._resource == 'environment':
            self.merge(mappings, self.environment_mappings())
        mappings.update(self.load_file(self.resource_folder, 'mappings'))
        return mappings

    def merge(self, a, b, key_path=None):
        if key_path is None: key_path = []
        for key in b:
            if key in a:
                if isinstance(a[key], dict) and isinstance(b[key], dict):
                    self.merge(a[key], b[key], key_path + [str(key)])
            else:
                a[key] = b[key]
        return a

    @staticmethod
    def _normalize_path(value): # pragma: no cover
        """Normalize the specified path value returning the absolute
        path for it.

        :param str value: The path value to normalize
        :rtype: str

        """
        return path.abspath(path.normpath(value))

    @property
    def resource_folder(self):
        """Return the folder that contains the resource's configuration data

        :rtype: str

        """
        if self._resource_type in ['environment', 'service']:
            return path.join(self.base_path, STACK_FOLDERS[self._resource_type],
                             self._resource)
        return path.join(self.base_path, STACK_FOLDERS[self._resource_type])

    @staticmethod
    def validate_config_path(config_path):
        """Validate that the specified configuration path exists and
        contains at least some of the files or folders expected.

        :param str config_path: The path to validate
        :rtype: bool

        """
        paths_found = [path.exists(config_path)]
        paths_found += [path.exists(path.join(config_path, f)) for f in
                        CONFIG_FILES]
        return any(paths_found)

    @staticmethod
    def validate_environment(config_path, env):
        """Validate that the expected environment ``network.yaml``
        configuration file exists within the config path for the environment.

        :param str config_path: The base config path
        :param str env: The environment name
        :rtype: bool

        """
        if not env:
            return False
        return path.exists(path.join(config_path,
                                     STACK_FOLDERS['environment'],
                                     env, 'environment.yaml'))

    @staticmethod
    def validate_resource(config_path, resource_type, resource):
        """Validate the resource configuration directory exists

        :param str resource_type: The resource type specified
        :param str resource: The resource to validate
        :rtype: bool

        """
        if resource_type in ['environment', 'service']:
            return path.exists(path.join(config_path,
                                         '{}s'.format(resource_type),
                                         resource,
                                         '{}.yaml'.format(resource_type)))
        return path.exists(path.join(config_path,
                                     STACK_FOLDERS[resource_type],
                                     '{}.yaml'.format(resource)))



"""
Build Cloud Formation ELB stacks

"""
from formulary.builders import base
from formulary.resources import elb

DEFAULT_PORT = 80
DEFAULT_PROTOCOL = 'http'


class LoadBalancer(base.Builder):
    """Build ELB Load Balancer stack"""
    def __init__(self, config, name, service, elb_config, instance_ids,
                 security_group, subnets):
        """Create an Elastic Load Balancer stack

        :param formulary.builders.config.Config: builder configuration
        :param str name: The base name fo the ELB stack
        :param dict elb_config: ELB specific configuration
        :param list instance_ids: List of reference IDs for the instances
        :param str security_group: The reference ID for the security group
        :param list subnets: List of subnet ids

        """
        super(LoadBalancer, self).__init__(config, name)
        internal = elb_config.get('internal')
        self._add_resource(self.name,
                           elb.LoadBalancer(self.name,
                                            [{'Ref': i} for i in instance_ids],
                                            self._health_check(elb_config),
                                            self._listeners(elb_config),
                                            [{'Ref': security_group}], subnets,
                                            internal=internal))
        self._add_output('DNSName',
                         'The DNSName for {0}'.format(self.full_name),
                         {'Fn::GetAtt': [self.reference_id, 'DNSName']})
        self._add_output('HostedZoneId',
                         'The HostedZoneId for {0}'.format(self.full_name),
                         {'Fn::GetAtt': [self.reference_id,
                                         'CanonicalHostedZoneNameID']})
        self._add_tag_to_resources('Environment', config.environment)
        self._add_tag_to_resources('Service', service)

    @staticmethod
    def _create_listener(config):
        port = config.get('port', DEFAULT_PORT)
        protocol = config.get('protocol', DEFAULT_PROTOCOL)
        instance_port = config.get('instance_port', port)
        instance_protocol = config.get('instance_protocol', protocol)
        kwargs = {'port': port,
                  'protocol': protocol,
                  'instance_port': instance_port,
                  'instance_protocol': instance_protocol,
                  'ssl_certificate_id': config.get('ssl_certificate_id')}
        print('kwargs', kwargs)
        for key, value in [(k, v) for k, v in kwargs.items()]:
            if value is None:
                del kwargs[key]
        return elb.Listener(**kwargs)

    @staticmethod
    def _health_check(config):
        protocol = config.get('instance_protocol',
                              config.get('protocol', DEFAULT_PROTOCOL))
        port = config.get('instance_port', config.get('port', DEFAULT_PORT))
        kwargs = {'interval': config.get('interval'),
                  'target': '{0}:{1}{2}'.format(protocol.upper(), port,
                                                config.get('check', '')),
                  'timeout': config.get('timeout'),
                  'healthy': config.get('healthy'),
                  'unhealthy': config.get('unhealthy')}
        for key, value in [(k, v) for k, v in kwargs.items()]:
            if value is None:
                del kwargs[key]
        return elb.HeathCheck(**kwargs)

    def _listeners(self, config):
        values = []
        if config.get('listeners'):
            for listener in config['listeners']:
                values.append(self._create_listener(listener))
        else:
            values.append(self._create_listener(config))
        return values



"""
Cloud Formation EC2 Resources

AWS::EC2::DHCPOptions
AWS::EC2::EIP
AWS::EC2::Instance
AWS::EC2::InternetGateway
AWS::EC2::NetworkAcl
AWS::EC2::NetworkAclEntry
AWS::EC2::Route
AWS::EC2::RouteTable
AWS::EC2::SecurityGroup
AWS::EC2::Subnet
AWS::EC2::SubnetNetworkAclAssociation
AWS::EC2::SubnetRouteTableAssociation
AWS::EC2::Volume
AWS::EC2::VolumeAttachment
AWS::EC2::VPC
AWS::EC2::VPCDHCPOptionsAssociation
AWS::EC2::VPCGatewayAttachment

"""
from formulary.resources import base
from formulary import utils


class BlockDevice(base.Property):
    """The Amazon EC2 block device mapping property is an embedded property of
    the AWS::EC2::Instance resource.

    """
    def __init__(self, name, ebs=None, no_device=None, virtual_name=None):
        super(BlockDevice, self).__init__()
        self._values = {'DeviceName': name,
                        'Ebs': ebs,
                        'NoDevice': no_device or {},
                        'VirtualName': virtual_name}


class EIP(base.Resource):
    """The AWS::EC2::EIP resource allocates an Elastic IP (EIP) address and
    can, optionally, associate it with an Amazon EC2 instance.

    """
    def __init__(self, instance_id):
        super(EIP, self).__init__('AWS::EC2::EIP')
        self._properties['InstanceId'] = instance_id
        self._properties['Domain'] = 'vpc'


class DHCPOptions(base.Resource):
    """Creates a set of DHCP options for your VPC."""
    def __init__(self, domain_name, name_servers, ntp_servers):
        super(DHCPOptions, self).__init__('AWS::EC2::DHCPOptions')
        self._properties['DomainName'] = domain_name
        self._properties['DomainNameServers'] = name_servers
        self._properties['NtpServers'] = ntp_servers


class Instance(base.CPResource):
    """The AWS::EC2::Instance type creates an Amazon EC2 instance."""
    def __init__(self, name, ami, availability_zone, block_devices,
                 instance_type, subnet, security_group, user_data,
                 private_ip=None, ebs=True, dependency=None):
        super(Instance, self).__init__('AWS::EC2::Instance')
        self._name = name
        self._subnet = subnet
        nic = {
            'AssociatePublicIpAddress': True,
            'DeviceIndex': '0',
            'GroupSet': [security_group],
            'SubnetId': subnet
        }
        if private_ip:
            nic['PrivateIpAddress'] = private_ip
        self._properties = {
            'AvailabilityZone': availability_zone,
            'BlockDeviceMappings': block_devices,
            'DisableApiTermination': False,
            'EbsOptimized': False,
            'ImageId': ami,
            'InstanceType': instance_type,
            'KeyName': {'Fn::FindInMap': ['AWS', 'KeyName', 'Value']},
            'Monitoring': False,
            'NetworkInterfaces': [nic],
            'UserData': user_data}
        if ebs:
            self._properties['InstanceInitiatedShutdownBehavior'] = 'stop'
        if dependency:
            self.set_dependency({"Ref": dependency})

    @property
    def subnet(self):
        return self._subnet


class InternetGateway(base.Resource):
    def __init__(self):
        super(InternetGateway, self).__init__('AWS::EC2::InternetGateway')


class MountPoint(base.Property):
    """The EC2 MountPoint property is an embedded property of the
    AWS::EC2::Instance type.

    """
    def __init__(self, device, volume_id):
        super(MountPoint, self).__init__()
        self._values = {'Device': device, 'VolumeId': volume_id}


class NetworkACL(base.Resource):
    def __init__(self, vpc_name, vpc_id):
        super(NetworkACL, self).__init__('AWS::EC2::NetworkAcl')
        self._name = '{0}-acl'.format(vpc_name)
        self._properties['VpcId'] = vpc_id


class NetworkACLEntry(base.Resource):
    def __init__(self, acl_id, cidr_block, rule_number, action, egress, ports):
        super(NetworkACLEntry, self).__init__('AWS::EC2::NetworkAclEntry')
        protocol, from_port, to_port = utils.parse_port_value(ports, -1)
        self._properties['NetworkAclId'] = acl_id
        self._properties['CidrBlock'] = cidr_block
        self._properties['RuleNumber'] = rule_number
        self._properties['Protocol'] = protocol
        self._properties['RuleAction'] = action
        self._properties['Egress'] = egress
        self._properties['PortRange'] = {'From': from_port, 'To': to_port}


class Route(base.Resource):
    def __init__(self, route_table_id, cidr_block, gateway_id, depends_on):
        super(Route, self).__init__('AWS::EC2::Route')
        self._properties['RouteTableId'] = route_table_id
        self._properties['DestinationCidrBlock'] = cidr_block
        self._properties['GatewayId'] = gateway_id
        self._attributes['DependsOn'] = depends_on


class RouteTable(base.Resource):
    def __init__(self, vpc_id):
        super(RouteTable, self).__init__('AWS::EC2::RouteTable')
        self._properties['VpcId'] = vpc_id


class SecurityGroup(base.Resource):
    """Creates an Amazon EC2 security group"""

    def __init__(self, name, description, vpc, ingress):
        super(SecurityGroup, self).__init__('AWS::EC2::SecurityGroup')
        self._name = name
        self._properties['GroupDescription'] = description
        self._properties['SecurityGroupIngress'] = ingress
        self._properties['VpcId'] = vpc


class SecurityGroupIngress(base.Resource):
    """The AWS::EC2::SecurityGroupIngress resource adds an ingress rule to an
    Amazon EC2 or Amazon VPC security group.

    """
    tags = False

    def __init__(self, group_id, ip_protocol, from_port, to_port,
                 cidr=None, source_security_group_id=None):
        super(SecurityGroupIngress,
              self).__init__('AWS::EC2::SecurityGroupIngress')
        self._properties['GroupId'] = group_id
        self._properties['IpProtocol'] = ip_protocol
        self._properties['FromPort'] = from_port
        self._properties['ToPort'] = to_port
        self._properties['SourceSecurityGroupId'] = source_security_group_id
        self._properties['CidrIp'] = cidr


class SecurityGroupRule(base.Property):
    def __init__(self, protocol, from_port,
                 to_port=None,
                 cidr_addr=None,
                 source_id=None,
                 source_name=None,
                 source_owner=None):
        super(SecurityGroupRule, self).__init__()
        self._values = {'CidrIp': cidr_addr,
                        'FromPort': from_port,
                        'IpProtocol': protocol,
                        'SourceSecurityGroupId': source_id,
                        'SourceSecurityGroupName': source_name,
                        'SourceSecurityGroupOwnerId': source_owner,
                        'ToPort': to_port or from_port}


class Subnet(base.Resource):
    """Creates a subnet in an existing VPC."""
    def __init__(self, vpc_name, subnet, vpc_id, az, cidr_block):
        super(Subnet, self).__init__('AWS::EC2::Subnet')
        self._name = '{0}{1}-subnet'.format(vpc_name, subnet)
        self._properties['AvailabilityZone'] = az
        self._properties['CidrBlock'] = cidr_block
        self._properties['VpcId'] = vpc_id


class SubnetNetworkAclAssociation(base.Resource):
    """Associates a subnet with a network ACL."""
    def __init__(self, subnet_id, network_acl_id):
        super(SubnetNetworkAclAssociation,
              self).__init__('AWS::EC2::SubnetNetworkAclAssociation')
        self._properties['SubnetId'] = subnet_id
        self._properties['NetworkAclId'] = network_acl_id


class SubnetRouteTableAssociation(base.Resource):
    """Associates a subnet with a route table."""
    def __init__(self, subnet_id, route_table_id):
        super(SubnetRouteTableAssociation,
              self).__init__('AWS::EC2::SubnetRouteTableAssociation')
        self._properties['SubnetId'] = subnet_id
        self._properties['RouteTableId'] = route_table_id


class Volume(base.Resource):
    """The AWS::EC2::Volume type creates a new Amazon Elastic Block Store
    volume.

    """
    def __init__(self, availability_zone, iops, size,
                 tags=None, volume_type=None, encrypted=False, key_id=None,
                 snapshot_id=None):
        super(Volume, self).__init__('AWS::EC2::Volume')
        if encrypted and not key_id:
            raise ValueError('key_id must be set if encryption is enabled')
        self._properties['AvailabilityZone'] = availability_zone
        self._properties['Iops'] = iops
        self._properties['Size'] = size
        self._properties['KmsKeyId'] = key_id,
        self._properties['Encrypted'] = encrypted
        self._properties['SnapshotId'] = snapshot_id
        self._properties['Tags'] = tags
        self._properties['VolumeType'] = volume_type


class VolumeAttachment(base.Resource):
    """Attaches an Amazon EBS volume to a running instance and exposes it to
    the instance with the specified device name.

    """
    def __init__(self, device, instance_id, volume_id):
        super(VolumeAttachment, self).__init__('AWS::EC2::VolumeAttachment')
        self._properties['Device'] = device
        self._properties['InstanceId'] = instance_id
        self._properties['VolumeId'] = volume_id


class VPC(base.Resource):
    """Creates a Virtual Private Cloud (VPC) with the CIDR block that you
    specify.

    """
    def __init__(self, name, dns_support, dns_hostnames, cidr_block):
        super(VPC, self).__init__('AWS::EC2::VPC')
        self._name = name
        self._properties['EnableDnsSupport'] = dns_support
        self._properties['EnableDnsHostnames'] = dns_hostnames
        self._properties['CidrBlock'] = cidr_block


class VPCDHCPOptionsAssociation(base.Resource):
    """Associates a set of DHCP options (that you've previously created) with
    the specified VPC.

    """
    def __init__(self, dhcp_id, vpc_id):
        super(VPCDHCPOptionsAssociation,
              self).__init__('AWS::EC2::VPCDHCPOptionsAssociation')
        self._properties = {'DhcpOptionsId': dhcp_id, 'VpcId': vpc_id}


class VPCGatewayAttachment(base.Resource):
    """Attaches a gateway to a VPC."""
    def __init__(self, gateway_id, vpc_id):
        super(VPCGatewayAttachment,
              self).__init__('AWS::EC2::VPCGatewayAttachment')
        self._properties = {'InternetGatewayId': gateway_id, 'VpcId': vpc_id}



"""
Base Resource and Property Classes

"""

class Property(object):
    """Cloud Formation Resource Property

    Represents a property of a Cloud Formation resource

    """
    def __init__(self):
        self._values = dict()

    def as_dict(self):
        values = dict(self._values)
        for key, value in self._values.items():
            if value is None or value == [] or value == {}:
                del values[key]
        return values


class Resource(object):
    """Cloud Formation Resource

    Represents a Cloud Formation resource as a class

    """
    tags = True

    def __init__(self, resource_type):
        """Create a new resource specifying the resource type

        :param str resource_type: Resource type such as ``AWS::EC2::Route``

        """
        self._attributes = {}
        self._dependency = None
        self._name = None
        self._properties = {}
        self._tags = {}
        self._type = resource_type

    def add_attribute(self, name, value):
        """Add a top-level attribute to the resource such as ``DependsOn`` in
        ``AWS::EC2::Route``

        :param str name: The attribute name
        :param str value: The attribute value

        """
        self._attributes[name] = value

    def add_property(self, name, value):
        """Add a property to the resource

        :param str name: The name of the property
        :param str|dict|list|bool|int value: The property value

        """
        self._properties[name] = value

    def add_tag(self, name, value):
        """Add a tag to the resource

        :param str name: The name of the tag
        :param str value: The tag value

        """
        self._tags[name] = value

    def as_dict(self):
        """Return the resource as a dictionary for building the cloud-formation
        configuration.

        :return: dict

        """
        self._prune_empty_properties()
        dict_val = dict({'Type': self._type, 'Properties': self._properties})
        if self._tags and self.tags is True:
            dict_val['Properties']['Tags'] = []
            for key, value in self._tags.items():
                dict_val['Properties']['Tags'].append({'Key': key,
                                                       'Value': value})
            if self._name:
                dict_val['Properties']['Tags'].append({'Key': 'Name',
                                                       'Value': self._name})
        for key, value in self._attributes.items():
            dict_val[key] = value
        if self._dependency:
            dict_val['DependsOn'] = self._dependency
        if not dict_val['Properties']:
            del dict_val['Properties']
        return dict_val

    def set_dependency(self, dependency):
        """Set a dependency for the resource to be created

        :param str|dict dependency: The dependency name or reference

        """
        self._dependency = dependency

    def set_name(self, name):
        """Set the resource name for use as a tag

        :param str name: The resource name

        """
        self._name = name

    def _prune_empty_properties(self):
        for key, value in list(self._properties.items()):
            if self._properties[key] is None:
                del self._properties[key]


class CPResource(Resource):
    """CPResources are resources that support Creation Policies"""
    def __init__(self, resource_type):
        super(CPResource, self).__init__(resource_type)
        self._policy = {}

    def as_dict(self):
        """Return the resource as a dictionary for building the cloud-formation
        configuration.

        :return: dict

        """
        dict_val = super(CPResource, self).as_dict()
        if self._policy:
            dict_val['CreationPolicy'] = dict(self._policy)
        return dict_val

    def set_handle(self, handle):
        self._properties['Handle'] = handle

    def set_timeout(self, timeout):
        self._properties['Timeout'] = timeout

    def set_creation_policy(self, signal_count, timeout):
        """Set a dependency for the resource to be created

        :param str dependency: The dependency name or reference

        """
        self._policy = {'ResourceSignal':
                            {'Count': signal_count,
                             'Timeout': timeout}}



"""

"""
import logging
import math
from os import path
import random

from formulary.builders import base

from formulary.resources import cloudformation
from formulary.builders import ec2
from formulary.resources import ec2 as ec2_resources
from formulary.builders import elb
from formulary.builders import route53
from formulary import utils

LOGGER = logging.getLogger(__name__)

# Defaults
DEFAULT_BLOCK_DEVICE = {'ebs': True, 'capacity': 20, 'type': 'gp2'}
DEFAULT_BLOCK_DEVICES = {'/dev/xvda': DEFAULT_BLOCK_DEVICE}


class Service(base.Builder):

    def __init__(self, config, name, amis, local_path, environment_stack,
                 dependency=None, wait_handle=None, parent=None):
        super(Service, self).__init__(config, name)

        self._amis = amis
        self._instances = []
        self._local_path = local_path
        self._mappings = config.mappings
        self._mappings.update(environment_stack.mappings)
        self._dependency = dependency
        self._wait_handle = wait_handle
        self._environment_stack = environment_stack
        self._parent = parent
        self._security_group = self._add_security_group()
        self._maybe_add_security_group_ingress()
        self._add_instances()
        self._maybe_add_elbs()
        self._add_tag_to_resources('Environment', self._config.environment)
        self._add_tag_to_resources('Service', self._name)

    def _add_autobalanced_instances(self, settings):
        count = settings.get('instance-count', 1)
        subnets = self._get_subnets(count)
        handle, wait = None, None
        for index in range(0, count):
            subnet = subnets.pop(0)
            if not handle:
                handle = self._maybe_add_wait_handle(index)
            ref_id = self._add_instance('instance{0}'.format(index), subnet,
                                        settings, handle, wait)
            if not wait:
                wait = self._maybe_add_wait_condition(index, handle, ref_id)

    # Add same-az instances in a method matching auto-balanced ones

    def _add_elb(self, name, config):
        if self._parent:
            name = '{0}-{1}'.format(self._parent, name)
        builder = elb.LoadBalancer(self._config, name, self.name, config,
                                   self._instances, 'SecurityGroupId',
                                   self.subnet_ids)
        builder.add_parameter('SecurityGroupId', {'Type': 'String'})
        parameters = {'SecurityGroupId':
                          {'Fn::GetAtt': [self._security_group,
                                          'Outputs.SecurityGroupId']}}

        for instance in self._instances:
            builder.add_parameter(instance, {'Type': 'String'})
            parameters[instance] = {'Fn::GetAtt': [instance,
                                                   'Outputs.InstanceId']}

        template_id, url = builder.upload(self._name)
        self._add_stack(name, url, parameters)
        self._maybe_add_route53_alias(config, utils.camel_case(name))

    def _add_instance(self, name, subnet, config,
                      wait_handle=None, dependency=None):
        """Add an instance to the resources for the given name, subnet, and
        config.

        :param str name: The name of the instance
        :param formulary.records.Subnet subnet: The subnet config
        :param dict config: The instance config
        :rtype: str

        """
        if self._parent:
            name = '{0}-{1}'.format(self._parent, name)
        LOGGER.debug('Adding %s', name)
        block_devices = self._get_block_devices(config.get('block_devices'))
        instance = ec2.Instance(self._config, name,
                                self._get_ami_id(),
                                block_devices,
                                config.get('instance-type'),
                                config.get('private_ip', ''),
                                {'Ref': 'SecurityGroupId'},
                                subnet,
                                self._read_user_data(),
                                self._tags,
                                config.get('ebs', True),
                                self._parent or self._environment_stack.name)

        instance.add_parameter('SecurityGroupId',
                               {'Type': 'String',
                                'Description': 'Security Group Physical ID'})

        parameters = {'SecurityGroupId':
                          {'Fn::GetAtt': [self._security_group,
                                          'Outputs.SecurityGroupId']}}

        wait_handle = wait_handle or self._wait_handle
        if wait_handle:
            instance.add_parameter(wait_handle, {'Type': 'String'})
            parameters[wait_handle] = {'Ref': wait_handle}

        # Add any SecurityGroupIngress resources for public-ingress support
        for port in config.get('public-ingress', []):
            protocol, from_port, to_port = utils.parse_port_value(port)
            cidr = {'Fn::Join': ['',
                                 [{'Fn::GetAtt': [utils.camel_case(name),
                                                  'PublicIp']}, '/32']]}
            resource = \
                ec2_resources.SecurityGroupIngress({'Ref': 'SecurityGroupId'},
                                                   protocol, from_port, to_port,
                                                   cidr)
            instance.add_resource('{0}-{1}-{2}'.format(self._name, protocol,
                                                       to_port), resource)

        template_id, url = instance.upload(self._name)

        dependency = dependency or config.get('dependency') or self._dependency
        self._add_stack(name, url, parameters, dependency=dependency)
        ref_id = utils.camel_case(name)
        self._instances.append(ref_id)
        return ref_id

    def _add_instances(self):
        settings = self._config.settings
        if 'instances' in settings:
            LOGGER.debug('Adding instances')
            for name, instance_cfg in settings['instances'].items():
                cfg = dict(settings)
                del cfg['instances']
                cfg.update(instance_cfg)
                for key in ['availability_zone', 'private_ip']:
                    cfg[key] = self._maybe_replace_with_mapping(cfg[key])
                self._maybe_add_availability_zone(cfg)
                self._add_instance(name,
                                   self._get_subnet(cfg['availability_zone']),
                                   cfg)
        elif settings.get('instance-strategy') == 'same-az':
            self._maybe_add_availability_zone(settings)
            subnet = self._get_subnet(settings['availability_zone'])
            for index in range(0, settings.get('instance-count', 1)):
                self._add_instance('instance{0}'.format(index),
                                   subnet, settings)
        elif settings.get('instance-strategy') == 'az-balanced':
            return self._add_autobalanced_instances(settings)
        elif 'instance-strategy' in settings:
            raise ValueError('Unknown instance-strategy: '
                             '{0}'.format(settings['instance-strategy']))

    def _add_security_group(self):
        LOGGER.debug('Adding security group')
        name = '{0}-security-group'.format(self._config.service)
        if self._parent:
            name = '{0}-security-group'.format(self._parent)
        builder = ec2.SecurityGroup(self._config, name,
                                    self._environment_stack, self._name)
        template_id, url = builder.upload(self._name)
        self._add_stack(name, url)
        return utils.camel_case(name)

    def _add_wait_condition(self, name):
        self._add_resource(name, cloudformation.WaitConditionHandle())

    def _get_ami_id(self):
        try:
            return self._amis[self._config.region][self._config.settings['ami']]
        except KeyError:
            raise ValueError('AMI %s not found' % self._config.settings['ami'])

    @staticmethod
    def _get_block_devices(devices):
        if devices == 'instance-store':
            return
        if not devices:
            devices = DEFAULT_BLOCK_DEVICES
        values = []
        for index, (device, config) in enumerate(devices.items()):
            kwargs = {'name': device}
            if config.get('ebs', DEFAULT_BLOCK_DEVICE['ebs']):
                kwargs['ebs'] = {
                    'VolumeType': config.get('type',
                                             DEFAULT_BLOCK_DEVICE['type']),
                    'VolumeSize': config.get('capacity',
                                             DEFAULT_BLOCK_DEVICE['capacity'])}
            else:
                kwargs['virtual_name'] = 'ephemeral{0}'.format(index)
            values.append(ec2_resources.BlockDevice(**kwargs).as_dict())

        LOGGER.debug('Block devices: %r', values)
        return values

    def _get_subnet(self, availability_zone):
        for subnet in self._get_subnets(len(self._environment_stack.subnets)):
            if subnet.availability_zone == availability_zone:
                return subnet

    def _get_subnets(self, count):
        subnets = self._environment_stack.subnets
        if count > len(subnets):
            subnets *= math.ceil(count / float(len(subnets)))
        return subnets[0:count]

    def _maybe_add_availability_zone(self, config):
        if 'availability_zone' not in config:
            offset = random.randint(0, len(self._environment_stack.subnets) - 1)
            config['availability_zone'] = \
                self._environment_stack.subnets[offset].availability_zone

    def _maybe_add_elbs(self):
        if not self._config.settings.get('elb'):
            return
        if isinstance(self._config.settings['elb'], list):
            for config in self._config.settings['elb']:
                self._add_elb(config['name'], config)
        elif isinstance(self._config.settings['elb'], dict):
            self._add_elb('elb', self._config.settings['elb'])

    def _maybe_add_route53_record(self, config):
        if 'route53' not in config:
            return
        name = '{0}-route53'.format(config['route53']['hostname'])
        builder = route53.Route53RecordSet(self._config, name,
                                           config['route53'], self._instances)
        template_id, url = builder.upload(self._name)
        params = dict()
        for instance in self._instances:
            params[instance] = {'Ref': instance}
        self._add_stack(name, url, params)

    def _maybe_add_route53_alias(self, config, ref_name):
        if 'route53' not in config:
            return
        name = '{0}-route53'.format(utils.dash_delimited(ref_name))
        builder = route53.Route53RecordSet(self._config, name,
                                           config['route53'])
        template_id, url = builder.upload(self._name)

        params = {'DNSName': {'Fn::GetAtt': [ref_name, 'Outputs.DNSName']},
                  'HostedZoneId': {'Fn::GetAtt': [ref_name,
                                                  'Outputs.HostedZoneId']}}

        self._add_stack(name, url, params)

    def _maybe_add_security_group_ingress(self):
        LOGGER.debug('Possibly adding security group ingress rules')
        name = 'security-group-ingress'
        if self._parent:
            name = '{0}-{1}'.format(self._parent, name)
        builder = ec2.SecurityGroupIngress(self._config, name,
                                           self._environment_stack,
                                           self._name)
        parameters = {'SecurityGroupId':
                          {'Fn::GetAtt': [self._security_group,
                                          'Outputs.SecurityGroupId']}}
        if builder.resources:
            template_id, url = builder.upload(self._name)
            self._add_stack(name, url, parameters,
                            dependency=self._security_group)

    def _maybe_add_wait_condition(self, index, handle, ref_id):
        if 'wait-condition' not in self._config.settings:
            return
        settings = self._config.settings['wait-condition']
        if index == settings['after-node']:
            name = '{0}-wait'.format(self._name)
            if self._parent:
                name = '{0}-{1}'.format(self._parent, name)
            handle = {'Ref': handle}
            timeout = settings.get('timeout', 3600)
            wait = cloudformation.WaitCondition(1, handle, timeout)
            wait.set_dependency(ref_id)
            self._add_resource(name, wait)
            cc_name = utils.camel_case(name)
            self._add_output(cc_name + 'Data', 'WaitCondition return data',
                             {'Fn::GetAtt': [cc_name, 'Data']})
            return cc_name

    def _maybe_add_wait_handle(self, index):
        if 'wait-condition' not in self._config.settings:
            return
        settings = self._config.settings['wait-condition']
        if index == settings['after-node']:
            name = utils.dash_delimited(settings['handle']) or self._name
            if self._parent:
                name = '{0}-{1}'.format(self._parent, name)
            self._add_resource(name,
                               cloudformation.WaitConditionHandle())
            return utils.camel_case(name)

    def _read_user_data(self):
        if self._config.settings.get('user-data'):
            with open(path.join(self._local_path,
                                self._config.settings['user-data'])) as handle:
                return handle.read()

    @property
    def subnet_ids(self):
        """Return a list of Subnet IDs for the VPC

        :rtype: list

        """
        return [s.id for s in self._environment_stack.subnets]



"""
Build the AWS VPC environment by adding the various resources to the
Cloud Formation template

"""

from formulary.builders import base
from formulary.resources import ec2
from formulary import utils


class Environment(base.Builder):

    def __init__(self, config, name):
        super(Environment, self).__init__(config, name)
        self._vpc, self._vpc_name = self._add_vpc()
        self._add_dhcp()
        self._gateway = self._add_gateway()
        self._internet_gateway = self._add_gateway_attachment()
        self._route_table = self._add_route_table()
        self._add_public_route()
        self._acl = self._add_network_acl()
        self._add_network_acl_entries()
        self._add_subnets()
        self._add_output('VPCId', 'VPC ID for {0}'.format(self._vpc_name),
                         self._vpc)

    def _add_vpc(self):
        """Add the VPC section to the template, returning the id and name
        for use in the other sections of the stack configuration

        :rtype: str, str

        """
        vpc_name = self._name.replace('_', '-')
        resource = ec2.VPC(vpc_name,
                           self._config.settings['vpc']['dns-support'],
                           self._config.settings['vpc']['dns-hostnames'],
                           self._config.settings['CIDR'])
        resource.add_tag('Environment', self._config.settings['environment'])
        return self._add_resource(vpc_name, resource), vpc_name

    def _add_dhcp(self):
        """Add all of the DHCP options to the template for the given VPC"""
        self._add_dhcp_association(self._add_dhcp_options())

    def _add_dhcp_options(self):
        """Add all of the DHCP options to the template for the given VPC

        :rtype: str

        """
        config = self._config.settings['dhcp-options']
        options = '{0}-dhcp'.format(self._vpc_name)
        self._add_resource(options,
                           ec2.DHCPOptions(config['domain-name'],
                                           config['name-servers'],
                                           config['ntp-servers']))
        return options

    def _add_dhcp_association(self, dhcp_id):
        """Add all of the DHCP OptionsAssociation to the template for the given
        VPC and DHCP Options ID.

        :param str dhcp_id: The DHCP Options ID

        """
        dhcp = {'Ref': utils.camel_case(dhcp_id)}
        vpc_id = {'Ref': utils.camel_case(self._vpc_name)}
        self._add_resource('{0}-dhcp-assoc'.format(dhcp_id),
                           ec2.VPCDHCPOptionsAssociation(dhcp, vpc_id))

    def _add_gateway(self):
        """Add a gateway to the template for the specified VPC

        :rtype: str

        """
        gateway = '{0}-gateway'.format(self._vpc_name)
        self._add_resource(gateway, ec2.InternetGateway())
        return gateway

    def _add_gateway_attachment(self):
        """Attach the specified gateway to the VPC

        :rtype: str

        """
        attachment = '{0}-attachment'.format(self._gateway)
        gateway_id = {'Ref': utils.camel_case(self._gateway)}
        vpc_id = {'Ref': utils.camel_case(self._vpc_name)}
        self._add_resource(attachment,
                           ec2.VPCGatewayAttachment(gateway_id, vpc_id))
        return attachment

    def _add_network_acl(self):
        """Add the Network ACL to the VPC

        :rtype: str

        """
        resource = ec2.NetworkACL(self._vpc_name,
                                  {'Ref': utils.camel_case(self._vpc_name)})
        resource.add_tag('Environment', self._config.settings['environment'])
        acl = '{0}-network-acl'.format(self._vpc_name)
        self._add_resource(acl, resource)
        return acl

    def _add_network_acl_entries(self):
        """Iterate through the ACL entries and add them"""
        acl_id = {'Ref': utils.camel_case(self._acl)}
        for index, acl in enumerate(self._config.settings['network-acls']):
            self._add_resource('{0}{1}'.format(self._acl, index),
                               ec2.NetworkACLEntry(acl_id,
                                                   acl['CIDR'],
                                                   acl['number'], acl['action'],
                                                   acl['egress'], acl['ports']))

    def _add_public_route(self):
        """Add the public route specified in the mapping ``pubic/cidr`` for
        the specified VPC, route table, gateway and internet gateway.

        """
        route_table_id = {'Ref': utils.camel_case(self._route_table)}
        gateway_id = {'Ref': utils.camel_case(self._gateway)}
        internet_gateway_id = utils.camel_case(self._internet_gateway)
        self._add_resource('{0}-route'.format(self._vpc_name),
                           ec2.Route(route_table_id,
                                     {'Fn::FindInMap': ['SubnetConfig',
                                                        'Public',
                                                        'CIDR']},
                                     gateway_id, internet_gateway_id))

    def _add_route_table(self,):
        """Add the the route table for the specified VPC

        :rtype: str

        """
        route_table = '{0}-route-table'.format(self._vpc_name)
        vpc_name = {'Ref': utils.camel_case(self._vpc_name)}
        self._add_resource(route_table, ec2.RouteTable(vpc_name))
        return route_table

    def _add_subnets(self):
        """Add the network subnets for the specified VPC and route table"""
        subnet_ids = []
        route_table_id = {'Ref': utils.camel_case(self._route_table)}
        vpc_id = {'Ref': utils.camel_case(self._vpc_name)}

        for subnet in self._config.settings['subnets']:
            config = self._config.settings['subnets'][subnet]

            subnet_id = '{0}{1}-subnet'.format(self._vpc_name, subnet)

            subnet_ids.append(utils.camel_case(subnet_id))
            resource = ec2.Subnet(self._vpc_name, subnet,
                                  vpc_id,
                                  config['availability_zone'],
                                  config['CIDR'])

            resource.add_tag('Environment',
                             self._config.settings['environment'])

            self._add_resource(subnet_id, resource)
            subnet_ref = {'Ref': utils.camel_case(subnet_id)}

            self._add_resource('{0}-assoc'.format(subnet_id),
                               ec2.SubnetRouteTableAssociation(subnet_ref,
                                                               route_table_id))



"""
Build Cloud Formation EC2 stacks

"""
import json
import logging
import re

from formulary.builders import base
from formulary.resources import ec2
from formulary import s3
from formulary import utils

LOGGER = logging.getLogger(__name__)

DEFAULT_INSTANCE_TYPE = 't2.small'
USER_DATA_RE = re.compile(r'\{\^(?P<command>instance|map)\s(?P<key>[\w\.]+)\}')
USER_DATA_JSON = re.compile(r'(\{\"(?:Ref|Fn\::\w+)\"[\s\:]+'
                            r'(?:\".*?\"|\[.*?\])\})')


class Instance(base.Builder):

    def __init__(self, config, name, ami, block_devices, instance_type,
                 private_ip, security_group, subnet, user_data, tags,
                 ebs=True, stack_name=None, dependency=None):
        """Create a new EC2 instance builder

        :param formulary.builders.config.Config: builder configuration
        :param str name:
        :param str ami:
        :param list block_devices:
        :param str instance_type:
        :param str private_ip:
        :param str security_group:
        :param formulary.resources.ec2.Subnet subnet:
        :param str user_data:
        :param dict tags: Tags to add to the resource
        :param bool ebs: Is EBS backed
        :param str stack_name: The name of the parent stack for the instance

        """
        super(Instance, self).__init__(config, name)
        self._s3 = s3.S3(config.s3_bucket, config.s3_prefix, config.profile)

        # Build kwargs used for user-data template and ec2.Instance
        kwargs = {'name': self.full_name,
                  'ami': ami,
                  'availability_zone': subnet.availability_zone,
                  'block_devices': block_devices,
                  'dependency': dependency,
                  'environment': config.environment,
                  'instance_type': instance_type,
                  'private_ip': private_ip,
                  'ref_id': self.reference_id,
                  'region': config.region,
                  'service': config.service,
                  'subnet': subnet.id,
                  'security_group': security_group,
                  'stack': stack_name,
                  'ebs': ebs}

        kwargs['user_data'] = self._render_user_data(user_data, kwargs)

        # Remove the kwargs that don't get passed to ec2.Instance
        for key in ['environment', 'ref_id', 'region', 'service', 'stack']:
            del kwargs[key]

        resource = ec2.Instance(**kwargs)
        for key, value in tags.items():
            resource.add_tag(key, value)
        ref_id = self._add_resource(self.name, resource)

        # Add private and public IP output
        self._add_output('InstanceId',
                         'The logical ID for {0}'.format(self.full_name),
                         {'Ref': ref_id})
        self._add_output('PrivateIP',
                         'Private IP address for {0}'.format(self.full_name),
                         {'Fn::GetAtt': [ref_id, 'PrivateIp']})
        self._add_output('PublicIP',
                         'Public IP address for {0}'.format(self.full_name),
                         {'Fn::GetAtt': [ref_id, 'PublicIp']})

    def _render_user_data(self, content, kwargs):
        for match in USER_DATA_RE.finditer(content):
            if match.group(1) == 'map':
                value = self._config.mappings
                for key in str(match.group(2)).split('.'):
                    try:
                        value = value.get(key.strip())
                    except AttributeError:
                        LOGGER.warning('Error assigning user-data value '
                                       'to "%s", value does not exist', key)
                content = content.replace(match.group(0), value)
            elif match.group(1) == 'instance':
                content = content.replace(match.group(0),
                                          kwargs[match.group(2)])
            elif match.group(1) == 's3file':
                value = self._s3.fetch(str(match.group(2)))
                content = content.replace(match.group(0), value)

        data = []
        for line in content.split('\n'):
            matches = USER_DATA_JSON.findall(line)
            if not matches:
                data.append(line + '\n')
                continue

            parts = USER_DATA_JSON.split(line)
            for offset, part in enumerate(parts):
                if part in matches:
                    parts[offset] = json.loads(part)
            if isinstance(parts[-1], dict):
                parts.append('\n')
            else:
                parts[-1] += '\n'
            for part in parts:
                data.append(part)

        return {'Fn::Base64': {'Fn::Join': ['', data]}}



class SecurityGroup(base.Builder):

    def __init__(self, config, name, stack, owner):
        super(SecurityGroup, self).__init__(config, name)
        self._owner = owner
        self._name = name
        self._stack = stack
        self._add_security_group()
        self._add_output('SecurityGroupId',
                         'The physical ID for the security group',
                         {'Ref': self.reference_id})

        self._add_tag_to_resources('Environment', config.environment)
        self._add_tag_to_resources('Service', owner)

    def _add_security_group(self):
        if isinstance(self._config.settings.get('security-group'), str):
            return self._config.settings.get('security-group')

        desc = ('Security Group for the {0} '
                'service in {1}').format(self._owner.capitalize(),
                                         self._stack.environment.capitalize())
        resource = ec2.SecurityGroup(self._name, desc, self._stack.vpc.id,
                                     self._build_ingress_rules())
        self._add_resource(self._name, resource)

    def _build_ingress_rules(self):
        rules = []
        group = self._config.settings.get('security-group', {})
        ingress_rules = list(group.get('ingress', {}))
        for row in ingress_rules:
            try:
                port, source = dict(row).popitem()
            except KeyError:
                continue
            if source == 'security-group':
                continue
            protocol, from_port, to_port = utils.parse_port_value(port)
            cidr_block = utils.find_in_map(source)
            rules.append(ec2.SecurityGroupRule(protocol,
                                               from_port, to_port,
                                               cidr_block).as_dict())
        return rules


class SecurityGroupIngress(base.Builder):

    def __init__(self, config, name, stack, owner):
        super(SecurityGroupIngress, self).__init__(config, name)
        self._owner = owner
        self._name = name
        self._stack = stack
        self._add_parameter('SecurityGroupId', {'Type': 'String'})

        for port in self._get_ingress_ports():
            protocol, from_port, to_port = utils.parse_port_value(port)
            ref_id = {'Ref': 'SecurityGroupId'}
            resource = ec2.SecurityGroupIngress(ref_id, protocol, from_port,
                                                to_port, ref_id)
            name = '{0}-{1}-{2}'.format(self._name, protocol, to_port)
            self._add_resource(name, resource)

    def _get_ingress_ports(self):
        ports = []
        group = self._config.settings.get('security-group', {})
        for row in list(group.get('ingress', {})):
            try:
                port, source = dict(row).popitem()
            except KeyError:
                continue
            if source == 'security-group':
                ports.append(port)
        return ports



"""
Cloud Formation Resources

"""
from formulary.resources import base


class Stack(base.Resource):
    """The AWS::CloudFormation::Stack type nests a stack as a resource in a
    top-level template.

    You can add output values from a nested stack within the containing
    template. You use the GetAtt function with the nested stack's logical name
    and the name of the output value in the nested stack in the format
    Outputs.NestedStackOutputName.

    When you apply template changes to update a top-level stack, AWS
    CloudFormation updates the top-level stack and initiates an update to
    its nested stacks. AWS CloudFormation updates the resources of modified
    nested stacks, but does not update the resources of unmodified nested
    stacks.

    """
    tags = False

    def __init__(self, template_url, parameters=None, notifications=None,
                 timeout=None):
        super(Stack, self).__init__('AWS::CloudFormation::Stack')
        self._properties = {'NotificationARNs': notifications,
                            'Parameters': parameters,
                            'TemplateURL': template_url,
                            'TimeoutInMinutes': timeout}


class WaitCondition(base.CPResource):
    """You can use a wait condition for situations like the following:

    To coordinate stack resource creation with configuration actions that are
    external to the stack creation

    To track the status of a configuration process

    For these situations, we recommend that you associate a CreationPolicy
    attribute with the wait condition so that you don't have to use a wait
    condition handle. For more information and an example, see Creating Wait
    Conditions in a Template. If you use a CreationPolicy with a wait
    condition, do not specify any of the wait condition's properties.

    """
    tags = False

    def __init__(self, count=None, handle=None, timeout=None):
        super(WaitCondition,
              self).__init__('AWS::CloudFormation::WaitCondition')
        self._properties = {'Count': count,
                            'Handle': handle,
                            'Timeout': timeout}


class WaitConditionHandle(base.Resource):
    """The AWS::CloudFormation::WaitConditionHandle type has no properties.
    When you reference the WaitConditionHandle resource by using the Ref
    function, AWS CloudFormation returns a presigned URL. You pass this URL to
    applications or scripts that are running on your Amazon EC2 instances to
    send signals to that URL. An associated AWS::CloudFormation::WaitCondition
    resource checks the URL for the required number of success signals or for
    a failure signal.

    """
    tags = False

    def __init__(self):
        super(WaitConditionHandle,
              self).__init__('AWS::CloudFormation::WaitConditionHandle')
        self._properties = {}



"""
Build the resources for a RDS Instance

"""
from formulary.builders import base
from formulary.builders import ec2
from formulary.resources import rds
from formulary import utils

DEFAULT_BACKUP_RETENTION = 3
DEFAULT_INSTANCE_TYPE = 'db.t2.small'
DEFAULT_ENGINE_TYPE = 'postgres'
DEFAULT_ENGINE_VERSION = '9.4.1'
DEFAULT_ENGINE_PORT = 5432
DEFAULT_STORAGE_CAPACITY = 100
DEFAULT_IOPS = 1000


class RDS(base.Builder):

    def __init__(self, config, name, stack):
        super(RDS, self).__init__(config, name)
        self._stack = stack
        self._security_group = self._add_security_group()
        self._add_instance()

    def _add_instance(self):
        subnets = self._add_subnet_groups()
        if not self._config.settings.get('multi-az'):
            self._config.settings['availability_zone'] = \
                self._stack.subnets[0].availability_zone
        name = '{0}-{1}'.format(self._config.environment, self._name)
        resource = rds.DBInstance(name,
                                  self._config.settings,
                                  utils.camel_case(subnets),
                                  {'Ref': self._security_group})
        resource.add_tag('Environment', self._stack.environment)
        resource.add_tag('Service', self._name)
        self._add_resource(name, resource)

    def _add_security_group(self):
        name = '{0}-security-group'.format(self._name)
        builder = ec2.SecurityGroup(self._config, name, self._stack,
                                    self._name)
        self._resources += builder.resources
        return builder.reference_id

    def _add_subnet_groups(self):
        subnets = []
        for subnet in self._stack.subnets:
            subnets.append(subnet.id)
        name = '{0}-{1}-subnet-group'.format(self._config.environment,
                                             self._name)
        resource = rds.DBSubnetGroup(name, subnets)
        self._add_resource(name, resource)
        return name



"""
Build the resources for a RDS Instance

"""
from formulary.builders import base
from formulary.resources import route53


class Route53RecordSet(base.Builder):

    def __init__(self, config, name, settings, instances=None):
        super(Route53RecordSet, self).__init__(config, name)
        self._add_parameter('DNSName', {'Type': 'String'})
        self._add_parameter('HostedZoneId', {'Type': 'String'})
        if instances:
            self._add_instance_rr(settings, instances)
        else:
            self._add_alias_record(settings)

    def _add_alias_record(self, settings):
        alias = route53.Route53AliasTarget({'Ref': 'DNSName'},
                                           {'Ref': 'HostedZoneId'})
        self._add_resource('route53-{0}-a'.format(settings['hostname']),
                           route53.Route53RecordSet(settings['domain_name'],
                                                    settings['hostname'], None,
                                                    alias.as_dict(), 'A'))

    def _add_instance_rr(self, settings, instances):
        for instance in instances:
            self._add_parameter(instance, {'Type': 'String'})
        resources = [{'Ref': ref_id} for ref_id in instances]
        self._add_resource('route53-{0}-a'.format(settings['hostname']),
                           route53.Route53RecordSet(settings['domain_name'],
                                                    settings['hostname'],
                                                    resources, None, 'A'))



"""
Cloud Formation methods and interfaces

"""
import collections
import json
import logging
import uuid

from boto3 import session
from botocore import exceptions

from formulary import records
from formulary import s3
from formulary import utils

LOGGER = logging.getLogger(__name__)
MAX_TEMPLATE_SIZE = 51200

StackResource = collections.namedtuple('StackResource',
                                       ('id', 'type', 'name', 'status'))


class CloudFormation(object):
    """Class for interfacing with Cloud Formation and related APIs"""

    def __init__(self, profile, region, s3bucket_name, s3bucket_path):
        """Estimate the cost of the stack in EC2

        :param str region: The region name to create the stack in
        :param str bucket: The bucket to use for large templates
        :param str profile: The credentials profile to use

        """
        self._s3 = s3.S3(s3bucket_name, s3bucket_path, profile)

        self._session = session.Session(profile_name=profile,
                                        region_name=region)
        self._client = self._session.client('cloudformation')

    def create_stack(self, template, environment, service=None):
        """Create a stack in the specified region with the given template,
        returning the stack id.

        :param Template template: The template to use
        :param str environment: The environment to set in a stack tag
        :param str|None service: The service name to set in a stack tag
        :rtype: str

        """
        template_id = str(uuid.uuid4())
        url = self._s3.upload(template_id, template.as_json())
        tags = [{'Key': 'Environment', 'Value': environment}]
        if service:
            tags.append({'Key': 'Service', 'Value': service})
        try:
            result = self._client.create_stack(StackName=template.name,
                                               Tags=tags,
                                               TemplateURL=url)
        except exceptions.ClientError as error:
            self._s3.delete(template_id)
            raise RequestException(error)

        LOGGER.debug('Created stack ID: %r', result['StackId'])

        # Upload stack details for removing completed stacks
        stack_details = {'id': result['StackId'],
                         'name': template.name,
                         'environment': environment}
        self._s3.upload('stack.json', json.dumps(stack_details))

        return result['StackId']

    def update_stack(self, template):
        """Update a stack in the specified region with the given template.

        :param Template template: The template to use
        :raises: RequestException

        """
        template_id = str(uuid.uuid4())
        url = self._s3.upload(template_id, template.as_json())
        try:
            result = self._client.update_stack(StackName=template.name,
                                               TemplateURL=url)
        except exceptions.ClientError as error:
            self._s3.delete(template_id)
            raise RequestException(error)
        LOGGER.debug('Updated stack ID: %r', result['StackId'])


class _API(object):

    def __init__(self, region, profile):
        kwargs = {'profile_name': profile} if profile else {}
        LOGGER.debug('Connecting with profile: %r', kwargs)
        self._session = session.Session(profile_name=profile,
                                        region_name=region)
        self._cf = self._session.client('cloudformation')
        self._ec2 = self._session.resource('ec2')

    @property
    def cf(self):
        """Return the connection handle to the Cloud Formation API

        :rtype: botocore.client.CloudFormation

        """
        return self._cf

    @property
    def ec2(self):
        """Return the connection handle to the EC2 API

        :rtype: botocore.client.EC2

        """
        return self._ec2


class EnvironmentStack(object):

    def __init__(self, name, config, template=None, profile=None):
        self._config = config
        self._name = name
        self._profile = profile
        self._region = config.get('region')

        self._api = None
        self._description = None
        self._id = None
        self._resources = {}
        self._status = None
        if template:
            self._process_template(template)
        else:
            self._process_api_data()

    @property
    def cidr_block(self):
        """Return the CIDR block for the Stack's VPC

        :rtype: str

        """
        return self.vpc.cidr_block

    @property
    def environment(self):
        """Return the environment from the Stack's tags

        :rtype: str

        """
        return self.vpc.tags.get('Environment', self._name)

    @property
    def mappings(self):
        """Return mappings based upon values in the stack used for service
        configuration and user-data templates.

        :rtype: dict

        """
        vpc_data = dict()
        for key, value in self.vpc._asdict().items():
            if key == 'tags':
                continue
            cckey = utils.camel_case(key)
            if key == 'cidr_block':
                cckey = 'CIDR'
            if value is not None:
                vpc_data[cckey] = value
        return {
            'Network': {
                'Name': {'Value': self._name},
                'VPC': vpc_data,
                'AWS': {'Region': self._region}
            }
        }

    @property
    def name(self):
        """Return the stack name

        :rtype: str

        """
        return self._name

    @property
    def resources(self):
        """Return the resources associated with this stack

        :rtype: list

        """
        return self._resources

    @property
    def subnets(self):
        """Return a list of subnets associated with the stack

        :rtype: list of formulary.records.Subnet

        """
        return [r for r in self._resources if isinstance(r, records.Subnet)]

    @property
    def vpc(self):
        """Return the VPC if it is in the resources.

        :rtype: formulary.records.VPC

        """
        for resource in self._resources:
            if isinstance(resource, records.VPC):
                return resource

    def _build_api_create_function_map(self):
        return {
            'AWS::EC2::DHCPOptions': self._create_dhcp_options_tuple,
            'AWS::EC2::VPCDHCPOptionsAssociation':
                self._create_dhcp_options_assoc_tuple,
            'AWS::EC2::InternetGateway': self._create_internet_gateway_tuple,
            'AWS::EC2::NetworkAcl': self._create_network_acl_tuple,
            'AWS::EC2::NetworkAclEntry': self._create_network_acl_tuple_entry,
            'AWS::EC2::Route': self._create_route_tuple,
            'AWS::EC2::RouteTable': self._create_route_table_tuple,
            'AWS::EC2::Subnet': self._create_subnet_tuple,
            'AWS::EC2::SubnetRouteTableAssociation':
                self._create_subnet_route_table_association,
            'AWS::EC2::VPC': self._create_vpc_tuple,
            'AWS::EC2::VPCGatewayAttachment':
                self._create_gateway_association_tuple
        }

    def _create_dhcp_options_tuple(self, value):
        """Create the DHCPOptions tuple from the retrieved stack resource
        information

        :param value: The DHCPOptions stack entry
        :type value: dict
        :rtype: formulary.records.DHCPOptions

        """
        record = self._describe_dhcp_options(value['PhysicalResourceId'])
        cfg = dict()
        for cfg_value in record.dhcp_configurations:
            cfg[cfg_value['Key']] = cfg_value['Values']
        return records.DHCPOptions(value['PhysicalResourceId'],
                                   value['LogicalResourceId'],
                                   value['Timestamp'].isoformat(),
                                   value['ResourceStatus'],
                                   cfg['domain-name'][0],
                                   cfg['domain-name-servers'][0],
                                   cfg['ntp-servers'][0],
                                   record.tags)

    @staticmethod
    def _create_dhcp_options_assoc_tuple(value):
        """Create the DHCPOptionsAssociation tuple from the retrieved stack
        resource information

        :param value: The DHCPOptionsAssociation stack entry
        :type value: dict
        :rtype: formulary.records.DHCPOptionsAssociation

        """
        return records.DHCPOptionsAssociation(value['PhysicalResourceId'],
                                              value['LogicalResourceId'],
                                              value['Timestamp'].isoformat(),
                                              value['ResourceStatus'])

    @staticmethod
    def _create_gateway_association_tuple(value):
        """Create the GatewayAssociation tuple from the retrieved stack
        resource information

        :param value: The gateway association stack entry
        :type value: dict
        :rtype: formulary.records.GatewayAssociation

        """
        return records.GatewayAssociation(value['PhysicalResourceId'],
                                          value['LogicalResourceId'],
                                          value['Timestamp'].isoformat(),
                                          value['ResourceStatus'])

    @staticmethod
    def _create_internet_gateway_tuple(value):
        """Create the InternetGateway tuple from the retrieved stack resource
        information

        :param value: The internet gateway stack entry
        :type value: dict
        :rtype: formulary.records.InternetGateway

        """
        return records.InternetGateway(value['PhysicalResourceId'],
                                       value['LogicalResourceId'],
                                       value['Timestamp'].isoformat(),
                                       value['ResourceStatus'])

    def _create_network_acl_tuple(self, value):
        """Create the NetworkACL tuple from the retrieved stack resource
        information

        :param value: The NetworkACL stack entry
        :type value: dict
        :rtype: formulary.records.NetworkACL

        """
        acl_data = self._describe_network_acl(value['PhysicalResourceId'])
        entries = []
        for entry in acl_data.entries:
            from_port, to_port = None, None
            if entry.get('PortRange'):
                from_port = entry['PortRange'].Get('From')
                to_port = entry['PortRange'].Get('To')
            acl_entry = records.NetworkACLEntry(None, None, None,
                                                int(entry['RuleNumber']),
                                                entry['CidrBlock'],
                                                bool(entry['Egress']),
                                                entry['RuleAction'],
                                                int(entry['Protocol']),
                                                from_port, to_port)
            entries.append(acl_entry)
        return records.NetworkACL(value['PhysicalResourceId'],
                                  value['LogicalResourceId'],
                                  value['Timestamp'].isoformat(),
                                  value['ResourceStatus'],
                                  entries,
                                  acl_data.tags)

    @staticmethod
    def _create_network_acl_tuple_entry(value):
        """Create the NetworkACLEntry tuple from the retrieved stack resource
        information

        :param value: The Network ACL entry stack entry
        :type value: dict
        :rtype: formulary.records.NetworkACL

        """
        return records.NetworkACLEntry(value['PhysicalResourceId'],
                                       value['Timestamp'].isoformat(),
                                       value['ResourceStatus'],
                                       None, None, None, None, None, None, None)

    @staticmethod
    def _create_route_tuple(value):
        """Create the Route tuple from the retrieved stack resource
        information

        :param value: The route stack entry
        :type value: dict
        :rtype: formulary.records.Route

        """
        return records.Route(value['PhysicalResourceId'],
                             value['LogicalResourceId'],
                             value['Timestamp'].isoformat(),
                             value['ResourceStatus'])

    def _create_route_table_tuple(self, value):
        """Create the RouteTable tuple from the retrieved stack resource
        information

        :param value: The route table stack entry
        :type value: dict
        :rtype: formulary.records.RouteTable

        """
        data = self._describe_route_table(value['PhysicalResourceId'])
        associations = [d.subnet_id for d in data.associations.all()]
        return records.RouteTable(value['PhysicalResourceId'],
                                  value['LogicalResourceId'],
                                  value['Timestamp'].isoformat(),
                                  value['ResourceStatus'],
                                  associations,
                                  data.tags)

    def _create_subnet_tuple(self, value):
        """Create the Subnet tuple from the retrieved stack resource
        information

        :param value: The subnet stack entry
        :type value: dict
        :rtype: formulary.records.Subnet

        """
        data = self._describe_subnet(value['PhysicalResourceId'])
        return records.Subnet(value['PhysicalResourceId'],
                              value['LogicalResourceId'],
                              value['Timestamp'].isoformat(),
                              value['ResourceStatus'],
                              data.availability_zone,
                              data.cidr_block,
                              data.available_ip_address_count)

    @staticmethod
    def _create_subnet_route_table_association(value):
        """Create the SubnetRouteTableAssociation tuple from the retrieved
        stack resource information

        :param value: The route stack entry
        :type value: dict
        :rtype: formulary.records.SubnetRouteTableAssociation

        """
        return \
            records.SubnetRouteTableAssociation(value['PhysicalResourceId'],
                                                value['LogicalResourceId'],
                                                value['Timestamp'].isoformat(),
                                                value['ResourceStatus'])

    def _create_vpc_tuple(self, value):
        """Create the VPC tuple from the retrieved stack resource information

        :param value: The VPC stack entry
        :type value: boto3.resources.factory.ec2.Vpc
        :return: formulary.records.VPC

        """
        vpc = self._get_vpc_resource(value['PhysicalResourceId'])
        tags = dict()
        for tag in vpc.tags:
            tags[tag['Key']] = tag['Value']
        return records.VPC(value['PhysicalResourceId'],
                           value['LogicalResourceId'],
                           value['Timestamp'].isoformat(),
                           vpc.state if vpc else None,
                           None,
                           vpc.cidr_block if vpc else None,
                           vpc.is_default if vpc else None,
                           vpc.instance_tenancy if vpc else None,
                           vpc.dhcp_options_id if vpc else None,
                           tags)

    def _describe_dhcp_options(self, dhcp_id):
        """Get the data from AWS about the specified DHCP options

        :param str dhcp_id: The ID to lookup
        :rtype: boto3.resources.factory.ec2.DhcpOptions

        """
        LOGGER.debug('Fetching data for %s', dhcp_id)
        try:
            return self._api.ec2.DhcpOptions(dhcp_id)
        except exceptions.ClientError as error:
            raise RequestException(error)

    def _describe_internet_gateway(self, gateway_id):
        """Get the data from AWS about the specified internet gateway

        :param str gateway_id: The ID to lookup
        :rtype: boto3.resources.factory.ec2.InternetGateway

        """
        LOGGER.debug('Fetching data for %s', gateway_id)
        try:
            return self._api.ec2.InternetGateway(gateway_id)
        except exceptions.ClientError as error:
            raise RequestException(error)

    def _describe_route_table(self, route_table_id):
        """Get the data from AWS about the specified route table

        :param str route_table_id: The ID to lookup
        :rtype: boto3.resources.factory.ec2.RouteTable

        """
        LOGGER.debug('Fetching data for %s', route_table_id)
        try:
            return self._api.ec2.RouteTable(route_table_id)
        except exceptions.ClientError as error:
            raise RequestException(error)

    def _describe_stack(self):
        """Fetch the core stack configuration data

        :rtype: dict

        """
        LOGGER.debug('Fetching data for %s', self._name)
        try:
            response = self._api.cf.describe_stacks(StackName=self._name)
        except exceptions.ClientError as error:
            raise RequestException(error)
        return response['Stacks'][0]

    def _describe_stack_resources(self):
        """Fetch the resources for the stack

        :rtype: list

        """
        function_map = self._build_api_create_function_map()

        LOGGER.debug('Fetching stack resources')
        resources = []
        try:
            result = self._api.cf.describe_stack_resources(StackName=self._name)
        except exceptions.ClientError as error:
            raise RequestException(error)
        for row in result['StackResources']:
            try:
                resources.append(function_map[row['ResourceType']](row))
            except KeyError:
                LOGGER.debug('Row data: %r', row)
                raise NotImplementedError(row.resource_type)
        return resources

    def _describe_network_acl(self, acl_id):
        """Get the data from AWS about the network ACL

        :param str acl_id: The ID to lookup
        :rtype: boto3.resources.factory.ec2.NetworkAcl

        """
        LOGGER.debug('Fetching data for %s', acl_id)
        try:
            return self._api.ec2.NetworkAcl(acl_id)
        except exceptions.ClientError as error:
            raise RequestException(error)

    def _describe_subnet(self, subnet_id):
        """Fetch the subnet configuration data

        :rtype: boto3.resources.factory.ec2.Subnet

        """
        LOGGER.debug('Fetching data for %s', subnet_id)
        try:
            return self._api.ec2.Subnet(subnet_id)
        except exceptions.ClientError as error:
            raise RequestException(error)

    def _get_vpc_resource(self, vpc_id):
        """Get the data from AWS about the VPC

        :param str vpc_id: The ID to lookup
        :rtype: boto.vpc.vpc.VPC

        """
        LOGGER.debug('Fetching data for %s', vpc_id)
        try:
            return self._api.ec2.Vpc(vpc_id)
        except exceptions.ClientError as error:
            raise RequestException(error)

    def _process_api_data(self):
        """Build the stack resources from a combination of remote API calls
        to various AWS endpoints.

        """
        LOGGER.debug('Processing API data')
        self._api = _API(self._region, self._profile)
        stack = self._describe_stack()
        self._description = stack['Description']
        self._id = stack['StackId']
        self._status = stack['StackStatus']
        self._resources = self._describe_stack_resources()

    def _process_template(self, template):
        values = json.loads(template.as_json())
        self._description = values['Description']
        # @TODO add resource tuples



class RequestException(Exception):
    def __init__(self, error):
        self._message = error.response['Error']['Message']

    def __repr__(self):
        return '<{0} "{1}">'.format(self._message)

    def __str__(self):
        return self._message
